@@ -1,10 +1,11 @@
 /*
  * id3tag.c -- Write ID3 version 1 and 2 tags.
  *
- * Copyright (C) 2000 Don Melton.
+ * Copyright (C) 2000 Don Melton
+ * Copyright (C) 2011 Robert Hegemann
  *
  * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
+ * modify it under the terms of the GNU Library General Public
  * License as published by the Free Software Foundation; either
  * version 2 of the License, or (at your option) any later version.
  *
@@ -13,7 +14,7 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Library General Public License for more details.
  *
- * You should have received a copy of the GNU Lesser General Public
+ * You should have received a copy of the GNU Library General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
  */
@@ -28,7 +29,7 @@
  * NOTE: See http://id3.org/ for more information about ID3 tag formats.
  */
 
-/* $Id: id3tag.c,v 1.53 2008/04/12 18:18:06 robert Exp $ */
+/* $Id: id3tag.c,v 1.73 2011/10/17 23:15:45 robert Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include <config.h>
@@ -47,16 +48,10 @@
 char   *strchr(), *strrchr();
 # ifndef HAVE_MEMCPY
 #  define memcpy(d, s, n) bcopy ((s), (d), (n))
-#  define memmove(d, s, n) bcopy ((s), (d), (n))
 # endif
 #endif
 
 
-#ifdef _MSC_VER
-#define snprintf _snprintf
-#endif
-
-
 #include "lame.h"
 #include "machine.h"
 #include "encoder.h"
@@ -91,7 +86,7 @@ static const char *const genre_names[] = {
     "Chanson", "Opera", "Chamber Music", "Sonata", "Symphony", "Booty Bass",
     "Primus", "Porn Groove", "Satire", "Slow Jam", "Club", "Tango", "Samba",
     "Folklore", "Ballad", "Power Ballad", "Rhythmic Soul", "Freestyle", "Duet",
-    "Punk Rock", "Drum Solo", "Acappella", "Euro-House", "Dance Hall",
+    "Punk Rock", "Drum Solo", "A Cappella", "Euro-House", "Dance Hall",
     "Goa", "Drum & Bass", "Club-House", "Hardcore", "Terror", "Indie",
     "BritPop", "Negerpunk", "Polsk Punk", "Beat", "Christian Gangsta",
     "Heavy Metal", "Black Metal", "Crossover", "Contemporary Christian",
@@ -156,28 +151,69 @@ typedef enum MiscIDs { ID_TXXX = FRAME_ID('T', 'X', 'X', 'X')
 } MiscIDs;
 
 
+static int
+frame_id_matches(int id, int mask)
+{
+    int     result = 0, i, window = 0xff;
+    for (i = 0; i < 4; ++i, window <<= 8) {
+        int const mw = (mask & window);
+        int const iw = (id & window);
+        if (mw != 0 && mw != iw) {
+            result |= iw;
+        }
+    }
+    return result;
+}
+
+static int
+isFrameIdMatching(int id, int mask)
+{
+    return frame_id_matches(id, mask) == 0 ? 1 : 0;
+}
+
+static int
+test_tag_spec_flags(lame_internal_flags const *gfc, unsigned int tst)
+{
+    return (gfc->tag_spec.flags & tst) != 0u ? 1 : 0;
+}
+
+#if 0
+static void
+debug_tag_spec_flags(lame_internal_flags * gfc, const char* info)
+{
+    MSGF(gfc, "%s\n", info);
+    MSGF(gfc, "CHANGED_FLAG  : %d\n", test_tag_spec_flags(gfc, CHANGED_FLAG )); 
+    MSGF(gfc, "ADD_V2_FLAG   : %d\n", test_tag_spec_flags(gfc, ADD_V2_FLAG  )); 
+    MSGF(gfc, "V1_ONLY_FLAG  : %d\n", test_tag_spec_flags(gfc, V1_ONLY_FLAG )); 
+    MSGF(gfc, "V2_ONLY_FLAG  : %d\n", test_tag_spec_flags(gfc, V2_ONLY_FLAG )); 
+    MSGF(gfc, "SPACE_V1_FLAG : %d\n", test_tag_spec_flags(gfc, SPACE_V1_FLAG)); 
+    MSGF(gfc, "PAD_V2_FLAG   : %d\n", test_tag_spec_flags(gfc, PAD_V2_FLAG  )); 
+}
+#endif
 
 
 
 static int
-        id3v2_add_ucs2(lame_global_flags * gfp, int frame_id, char const *lang,
-                       unsigned short const *desc, unsigned short const *text);
+id3v2_add_ucs2(lame_internal_flags * gfc, uint32_t frame_id, char const *lang,
+               unsigned short const *desc, unsigned short const *text);
 static int
-        id3v2_add_latin1(lame_global_flags * gfp, int frame_id, char const *lang, char const *desc,
-                         char const *text);
+id3v2_add_latin1(lame_internal_flags * gfc, uint32_t frame_id, char const *lang, char const *desc,
+                 char const *text);
 
 static void
-copyV1ToV2(lame_global_flags * gfp, int frame_id, char const *s)
+copyV1ToV2(lame_internal_flags * gfc, int frame_id, char const *s)
 {
-    lame_internal_flags *gfc = gfp->internal_flags;
-    int     flags = gfc->tag_spec.flags;
-    id3v2_add_latin1(gfp, frame_id, 0, 0, s);
+    unsigned int flags = gfc->tag_spec.flags;
+    id3v2_add_latin1(gfc, frame_id, 0, 0, s);
     gfc->tag_spec.flags = flags;
+#if 0
+    debug_tag_spec_flags(gfc, "copyV1ToV2");
+#endif
 }
 
 
 static void
-id3v2AddLameVersion(lame_global_flags * gfp)
+id3v2AddLameVersion(lame_internal_flags * gfc)
 {
     char    buffer[1024];
     const char *b = get_lame_os_bitness();
@@ -191,19 +227,19 @@ id3v2AddLameVersion(lame_global_flags * gfp)
     else {
         sprintf(buffer, "LAME version %s (%s)", v, u);
     }
-    copyV1ToV2(gfp, ID_ENCODER, buffer);
+    copyV1ToV2(gfc, ID_ENCODER, buffer);
 }
 
 static void
-id3v2AddAudioDuration(lame_global_flags * gfp)
+id3v2AddAudioDuration(lame_internal_flags * gfc, double ms)
 {
+    SessionConfig_t const *const cfg = &gfc->cfg;
     char    buffer[1024];
     double const max_ulong = MAX_U_32_NUM;
-    double  ms = gfp->num_samples;
     unsigned long playlength_ms;
 
     ms *= 1000;
-    ms /= gfp->in_samplerate;
+    ms /= cfg->samplerate_in;
     if (ms > max_ulong) {
         playlength_ms = max_ulong;
     }
@@ -213,8 +249,8 @@ id3v2AddAudioDuration(lame_global_flags * gfp)
     else {
         playlength_ms = ms;
     }
-    snprintf(buffer, sizeof(buffer), "%lu", playlength_ms);
-    copyV1ToV2(gfp, ID_PLAYLENGTH, buffer);
+    sprintf(buffer, "%lu", playlength_ms);
+    copyV1ToV2(gfc, ID_PLAYLENGTH, buffer);
 }
 
 void
@@ -233,6 +269,8 @@ id3tag_genre_list(void (*handler) (int, const char *, void *), void *cookie)
 
 #define GENRE_NUM_UNKNOWN 255
 
+
+
 void
 id3tag_init(lame_global_flags * gfp)
 {
@@ -240,7 +278,8 @@ id3tag_init(lame_global_flags * gfp)
     free_id3tag(gfc);
     memset(&gfc->tag_spec, 0, sizeof gfc->tag_spec);
     gfc->tag_spec.genre_id3v1 = GENRE_NUM_UNKNOWN;
-    id3v2AddLameVersion(gfp);
+    gfc->tag_spec.padding_size = 128;
+    id3v2AddLameVersion(gfc);
 }
 
 
@@ -279,10 +318,27 @@ id3tag_space_v1(lame_global_flags * gfp)
 
 void
 id3tag_pad_v2(lame_global_flags * gfp)
+{
+    id3tag_set_pad(gfp, 128);
+}
+
+void
+id3tag_set_pad(lame_global_flags * gfp, size_t n)
 {
     lame_internal_flags *gfc = gfp->internal_flags;
     gfc->tag_spec.flags &= ~V1_ONLY_FLAG;
     gfc->tag_spec.flags |= PAD_V2_FLAG;
+    gfc->tag_spec.flags |= ADD_V2_FLAG;
+    gfc->tag_spec.padding_size = n;
+}
+
+static int
+hasUcs2ByteOrderMarker(unsigned short bom)
+{
+    if (bom == 0xFFFEu || bom == 0xFEFFu) {
+        return 1;
+    }
+    return 0;
 }
 
 
@@ -337,7 +393,7 @@ local_ucs2_strdup(unsigned short **dst, unsigned short const *src)
     return 0;
 }
 
-#if 0
+
 static  size_t
 local_ucs2_strlen(unsigned short const *s)
 {
@@ -349,7 +405,43 @@ local_ucs2_strlen(unsigned short const *s)
     }
     return n;
 }
-#endif
+
+
+static size_t
+local_ucs2_substr(unsigned short** dst, unsigned short const* src, size_t start, size_t end)
+{
+    size_t const len = 1 + 1 + ((start < end) ? (end - start) : 0);
+    size_t n = 0;
+    unsigned short *ptr = calloc(len, sizeof(ptr[0]));
+    *dst = ptr;
+    if (ptr == 0 || src == 0) {
+        return 0;
+    }
+    if (hasUcs2ByteOrderMarker(src[0])) {
+        ptr[n++] = src[0];
+        if (start == 0) {
+            ++start;
+        }
+    }
+    while (start < end) {
+        ptr[n++] = src[start++];
+    }
+    ptr[n] = 0;
+    return n;
+}
+
+static int
+local_ucs2_pos(unsigned short const* str, unsigned short c)
+{
+    int     i;
+    for (i = 0; str != 0 && str[i] != 0; ++i) {
+        if (str[i] == c) {
+            return i;
+        }
+    }
+    return -1;
+}
+
 
 /*
 Some existing options for ID3 tag can be specified by --tv option
@@ -363,16 +455,12 @@ as follows.
 (although some are not exactly same)*/
 
 int
-id3tag_set_albumart(lame_global_flags * gfp, const char *image, unsigned long size)
+id3tag_set_albumart(lame_global_flags * gfp, const char *image, size_t size)
 {
     int     mimetype = 0;
     unsigned char const *data = (unsigned char const *) image;
     lame_internal_flags *gfc = gfp->internal_flags;
 
-    /* make sure the image size is no larger than the maximum value */
-    if (LAME_MAXALBUMART < size) {
-        return -1;
-    }
     /* determine MIME type from the actual image data */
     if (2 < size && data[0] == 0xFF && data[1] == 0xD8) {
         mimetype = MIMETYPE_JPEG;
@@ -411,16 +499,16 @@ set_4_byte_value(unsigned char *bytes, uint32_t value)
 {
     int     i;
     for (i = 3; i >= 0; --i) {
-        bytes[i] = value & 0xfful;
+        bytes[i] = value & 0xffUL;
         value >>= 8;
     }
     return bytes + 4;
 }
 
-static int
+static uint32_t
 toID3v2TagId(char const *s)
 {
-    int     i, x = 0;
+    unsigned int i, x = 0;
     if (s == 0) {
         return 0;
     }
@@ -438,8 +526,30 @@ toID3v2TagId(char const *s)
     return x;
 }
 
+static uint32_t
+toID3v2TagId_ucs2(unsigned short const *s)
+{
+    unsigned int i, x = 0;
+    if (s == 0) {
+        return 0;
+    }
+    for (i = 0; i < 4 && s[i] != 0; ++i) {
+        unsigned short const c = s[i];
+        unsigned int const u = 0x0ff & c;
+        x <<= 8;
+        x |= u;
+        if (c < 'A' || 'Z' < c) {
+            if (c < '0' || '9' < c) {
+                return 0;
+            }
+        }
+    }
+    return x;
+}
+
+#if 0
 static int
-isNumericString(int frame_id)
+isNumericString(uint32_t frame_id)
 {
     switch (frame_id) {
     case ID_DATE:
@@ -451,9 +561,10 @@ isNumericString(int frame_id)
     }
     return 0;
 }
+#endif
 
 static int
-isMultiFrame(int frame_id)
+isMultiFrame(uint32_t frame_id)
 {
     switch (frame_id) {
     case ID_TXXX:
@@ -486,17 +597,8 @@ isFullTextString(int frame_id)
 }
 #endif
 
-static int
-hasUcs2ByteOrderMarker(unsigned short bom)
-{
-    if (bom == 0xFFFEu || bom == 0xFEFFu) {
-        return 1;
-    }
-    return 0;
-}
-
 static FrameDataNode *
-findNode(id3tag_spec const *tag, int frame_id, FrameDataNode * last)
+findNode(id3tag_spec const *tag, uint32_t frame_id, FrameDataNode const *last)
 {
     FrameDataNode *node = last ? last->nxt : tag->v2_head;
     while (node != 0) {
@@ -561,7 +663,7 @@ isSameLang(char const *l1, char const *l2)
 }
 
 static int
-isSameDescriptor(FrameDataNode * node, char const *dsc)
+isSameDescriptor(FrameDataNode const *node, char const *dsc)
 {
     size_t  i;
     if (node->dsc.enc == 1 && node->dsc.dim > 0) {
@@ -591,10 +693,9 @@ isSameDescriptorUcs2(FrameDataNode const *node, unsigned short const *dsc)
 }
 
 static int
-id3v2_add_ucs2(lame_global_flags * gfp, int frame_id, char const *lang, unsigned short const *desc,
-               unsigned short const *text)
+id3v2_add_ucs2(lame_internal_flags * gfc, uint32_t frame_id, char const *lang,
+               unsigned short const *desc, unsigned short const *text)
 {
-    lame_internal_flags *gfc = gfp->internal_flags;
     if (gfc != 0) {
         FrameDataNode *node = 0;
         node = findNode(&gfc->tag_spec, frame_id, 0);
@@ -627,10 +728,9 @@ id3v2_add_ucs2(lame_global_flags * gfp, int frame_id, char const *lang, unsigned
 }
 
 static int
-id3v2_add_latin1(lame_global_flags * gfp, int frame_id, char const *lang, char const *desc,
+id3v2_add_latin1(lame_internal_flags * gfc, uint32_t frame_id, char const *lang, char const *desc,
                  char const *text)
 {
-    lame_internal_flags *gfc = gfp->internal_flags;
     if (gfc != 0) {
         FrameDataNode *node = 0;
         node = findNode(&gfc->tag_spec, frame_id, 0);
@@ -663,18 +763,54 @@ id3v2_add_latin1(lame_global_flags * gfp, int frame_id, char const *lang, char c
 }
 
 
-int
-id3tag_set_textinfo_ucs2(lame_global_flags * gfp, char const *id, unsigned short const *text)
+static int
+id3tag_set_userinfo_latin1(lame_internal_flags* gfc, uint32_t id, char const *fieldvalue)
 {
-    int const t_mask = FRAME_ID('T', 0, 0, 0);
-    int const frame_id = toID3v2TagId(id);
+    int     rc;
+    char* dsc = 0, *val;
+    local_strdup(&dsc, fieldvalue);
+    val = dsc;
+    while (*val) {
+        if (*val == '=') {
+            *val++ = 0;
+            break;
+        }
+        ++val;
+    }
+    rc = id3v2_add_latin1(gfc, id, "XXX", dsc, val);
+    free(dsc);
+    return rc;
+}
+
+static int
+id3tag_set_userinfo_ucs2(lame_internal_flags* gfc, uint32_t id, unsigned short const *fieldvalue)
+{
+    int     a, b, rc;
+    unsigned short* dsc = 0, *val = 0;
+    b = local_ucs2_strlen(fieldvalue);
+    a = local_ucs2_pos(fieldvalue, '=');
+    local_ucs2_substr(&dsc, fieldvalue, 0, a);
+    local_ucs2_substr(&val, fieldvalue, a+1, b);
+    rc = id3v2_add_ucs2(gfc, id, "XXX", dsc, val);
+    free(dsc);
+    free(val);
+    return rc;
+}
+
+int
+id3tag_set_textinfo_utf16(lame_global_flags * gfp, char const *id, unsigned short const *text)
+{
+    uint32_t const frame_id = toID3v2TagId(id);
     if (frame_id == 0) {
         return -1;
     }
-    if ((frame_id & t_mask) == t_mask) {
+    if (isFrameIdMatching(frame_id, FRAME_ID('T', 0, 0, 0))
+      ||isFrameIdMatching(frame_id, FRAME_ID('W', 0, 0, 0))) {
+#if 0
         if (isNumericString(frame_id)) {
             return -2;  /* must be Latin-1 encoded */
         }
+#endif
         if (text == 0) {
             return 0;
         }
@@ -682,26 +818,41 @@ id3tag_set_textinfo_ucs2(lame_global_flags * gfp, char const *id, unsigned short
             return -3;  /* BOM missing */
         }
         if (gfp != 0) {
-            return id3v2_add_ucs2(gfp, frame_id, 0, 0, text);
+            if (frame_id == ID_TXXX || frame_id == ID_WXXX) {
+                return id3tag_set_userinfo_ucs2(gfp->internal_flags, frame_id, text);
+            }
+            return id3v2_add_ucs2(gfp->internal_flags, frame_id, 0, 0, text);
         }
     }
     return -255;        /* not supported by now */
 }
 
+extern int
+id3tag_set_textinfo_ucs2(lame_global_flags * gfp, char const *id, unsigned short const *text);
+
+int
+id3tag_set_textinfo_ucs2(lame_global_flags * gfp, char const *id, unsigned short const *text)
+{
+    return id3tag_set_textinfo_utf16(gfp, id, text);
+}
+
 int
 id3tag_set_textinfo_latin1(lame_global_flags * gfp, char const *id, char const *text)
 {
-    int const t_mask = FRAME_ID('T', 0, 0, 0);
-    int const frame_id = toID3v2TagId(id);
+    uint32_t const frame_id = toID3v2TagId(id);
     if (frame_id == 0) {
         return -1;
     }
-    if ((frame_id & t_mask) == t_mask) {
+    if (isFrameIdMatching(frame_id, FRAME_ID('T', 0, 0, 0))
+      ||isFrameIdMatching(frame_id, FRAME_ID('W', 0, 0, 0))) {
         if (text == 0) {
             return 0;
         }
         if (gfp != 0) {
-            return id3v2_add_latin1(gfp, frame_id, 0, 0, text);
+            if (frame_id == ID_TXXX || frame_id == ID_WXXX) {
+                return id3tag_set_userinfo_latin1(gfp->internal_flags, frame_id, text);
+            }
+            return id3v2_add_latin1(gfp->internal_flags, frame_id, 0, 0, text);
         }
     }
     return -255;        /* not supported by now */
@@ -713,20 +864,32 @@ id3tag_set_comment_latin1(lame_global_flags * gfp, char const *lang, char const
                           char const *text)
 {
     if (gfp != 0) {
-        return id3v2_add_latin1(gfp, ID_COMMENT, lang, desc, text);
+        return id3v2_add_latin1(gfp->internal_flags, ID_COMMENT, lang, desc, text);
     }
     return -255;
 }
 
 
+int
+id3tag_set_comment_utf16(lame_global_flags * gfp, char const *lang, unsigned short const *desc,
+                         unsigned short const *text)
+{
+    if (gfp != 0) {
+        return id3v2_add_ucs2(gfp->internal_flags, ID_COMMENT, lang, desc, text);
+    }
+    return -255;
+}
+
+extern int
+id3tag_set_comment_ucs2(lame_global_flags * gfp, char const *lang, unsigned short const *desc,
+                        unsigned short const *text);
+
+
 int
 id3tag_set_comment_ucs2(lame_global_flags * gfp, char const *lang, unsigned short const *desc,
                         unsigned short const *text)
 {
-    if (gfp != 0) {
-        return id3v2_add_ucs2(gfp, ID_COMMENT, lang, desc, text);
-    }
-    return -255;
+    return id3tag_set_comment_utf16(gfp, lang, desc, text);
 }
 
 
@@ -737,7 +900,7 @@ id3tag_set_title(lame_global_flags * gfp, const char *title)
     if (title && *title) {
         local_strdup(&gfc->tag_spec.title, title);
         gfc->tag_spec.flags |= CHANGED_FLAG;
-        copyV1ToV2(gfp, ID_TITLE, title);
+        copyV1ToV2(gfc, ID_TITLE, title);
     }
 }
 
@@ -748,7 +911,7 @@ id3tag_set_artist(lame_global_flags * gfp, const char *artist)
     if (artist && *artist) {
         local_strdup(&gfc->tag_spec.artist, artist);
         gfc->tag_spec.flags |= CHANGED_FLAG;
-        copyV1ToV2(gfp, ID_ARTIST, artist);
+        copyV1ToV2(gfc, ID_ARTIST, artist);
     }
 }
 
@@ -759,7 +922,7 @@ id3tag_set_album(lame_global_flags * gfp, const char *album)
     if (album && *album) {
         local_strdup(&gfc->tag_spec.album, album);
         gfc->tag_spec.flags |= CHANGED_FLAG;
-        copyV1ToV2(gfp, ID_ALBUM, album);
+        copyV1ToV2(gfc, ID_ALBUM, album);
     }
 }
 
@@ -780,7 +943,7 @@ id3tag_set_year(lame_global_flags * gfp, const char *year)
             gfc->tag_spec.year = num;
             gfc->tag_spec.flags |= CHANGED_FLAG;
         }
-        copyV1ToV2(gfp, ID_YEAR, year);
+        copyV1ToV2(gfc, ID_YEAR, year);
     }
 }
 
@@ -792,8 +955,8 @@ id3tag_set_comment(lame_global_flags * gfp, const char *comment)
         local_strdup(&gfc->tag_spec.comment, comment);
         gfc->tag_spec.flags |= CHANGED_FLAG;
         {
-            int const flags = gfc->tag_spec.flags;
-            id3v2_add_latin1(gfp, ID_COMMENT, "XXX", "", comment);
+            uint32_t const flags = gfc->tag_spec.flags;
+            id3v2_add_latin1(gfc, ID_COMMENT, "XXX", "", comment);
             gfc->tag_spec.flags = flags;
         }
     }
@@ -802,7 +965,7 @@ id3tag_set_comment(lame_global_flags * gfp, const char *comment)
 int
 id3tag_set_track(lame_global_flags * gfp, const char *track)
 {
-    char   *trackcount;
+    char const *trackcount;
     lame_internal_flags *gfc = gfp->internal_flags;
     int     ret = 0;
 
@@ -823,7 +986,7 @@ id3tag_set_track(lame_global_flags * gfp, const char *track)
         if (trackcount && *trackcount) {
             gfc->tag_spec.flags |= (CHANGED_FLAG | ADD_V2_FLAG);
         }
-        copyV1ToV2(gfp, ID_TRACK, track);
+        copyV1ToV2(gfc, ID_TRACK, track);
     }
     return ret;
 }
@@ -847,6 +1010,72 @@ local_strcasecmp(const char *s1, const char *s2)
 }
 
 
+static 
+const char* nextUpperAlpha(const char* p, char x)
+{
+    char c;
+    for(c = toupper(*p); *p != 0; c = toupper(*++p)) {
+        if ('A' <= c && c <= 'Z') {
+            if (c != x) {
+                return p;
+            }
+        }
+    }
+    return p;
+}
+
+
+static int
+sloppyCompared(const char* p, const char* q)
+{
+    char cp, cq;
+    p = nextUpperAlpha(p, 0);
+    q = nextUpperAlpha(q, 0);
+    cp = toupper(*p);
+    cq = toupper(*q);
+    while (cp == cq) {
+        if (cp == 0) {
+            return 1;
+        }
+        if (p[1] == '.') { /* some abbrevation */
+            while (*q && *q++ != ' ') {
+            }
+        }
+        p = nextUpperAlpha(p, cp);
+        q = nextUpperAlpha(q, cq);
+        cp = toupper(*p);
+        cq = toupper(*q);
+    }
+    return 0;
+}
+
+
+static int 
+sloppySearchGenre(const char *genre)
+{
+    int i;
+    for (i = 0; i < GENRE_NAME_COUNT; ++i) {
+        if (sloppyCompared(genre, genre_names[i])) {
+            return i;
+        }
+    }
+    return GENRE_NAME_COUNT;
+}
+
+
+static int
+searchGenre(const char* genre)
+{
+    int i;
+    for (i = 0; i < GENRE_NAME_COUNT; ++i) {
+        if (!local_strcasecmp(genre, genre_names[i])) {
+            return i;
+        }
+    }
+    return GENRE_NAME_COUNT;
+}
+
+
 int
 id3tag_set_genre(lame_global_flags * gfp, const char *genre)
 {
@@ -857,17 +1086,17 @@ id3tag_set_genre(lame_global_flags * gfp, const char *genre)
         int     num = strtol(genre, &str, 10);
         /* is the input a string or a valid number? */
         if (*str) {
-            int     i;
-            for (i = 0; i < GENRE_NAME_COUNT; ++i) {
-                if (!local_strcasecmp(genre, genre_names[i])) {
-                    num = i;
-                    break;
-                }
+            num = searchGenre(genre);
+            if (num == GENRE_NAME_COUNT) {
+                num = sloppySearchGenre(genre);
             }
-            if (i == GENRE_NAME_COUNT) {
+            if (num == GENRE_NAME_COUNT) {
                 num = GENRE_INDEX_OTHER;
                 ret = -2;
             }
+            else {
+                genre = genre_names[num];
+            }
         }
         else {
             if ((num < 0) || (num >= GENRE_NAME_COUNT)) {
@@ -880,7 +1109,7 @@ id3tag_set_genre(lame_global_flags * gfp, const char *genre)
         if (ret) {
             gfc->tag_spec.flags |= ADD_V2_FLAG;
         }
-        copyV1ToV2(gfp, ID_GENRE, genre);
+        copyV1ToV2(gfc, ID_GENRE, genre);
     }
     return ret;
 }
@@ -919,9 +1148,15 @@ sizeOfNode(FrameDataNode const *node)
         switch (node->txt.enc) {
         default:
         case 0:
+            if (node->dsc.dim > 0) {
+                n += node->dsc.dim + 1;
+            }
             n += node->txt.dim;
             break;
         case 1:
+            if (node->dsc.dim > 0) {
+                n += (node->dsc.dim+1) * 2;
+            }
             n += node->txt.dim * 2;
             break;
         }
@@ -959,6 +1194,37 @@ sizeOfCommentNode(FrameDataNode const *node)
     return n;
 }
 
+static size_t
+sizeOfWxxxNode(FrameDataNode const *node)
+{
+    size_t  n = 0;
+    if (node) {
+        n = 10;         /* header size */
+        if (node->dsc.dim > 0) {
+            n += 1;         /* text encoding flag */
+            switch (node->dsc.enc) {
+            default:
+            case 0:
+                n += 1 + node->dsc.dim;
+                break;
+            case 1:
+                n += 2 + node->dsc.dim * 2;
+                break;
+            }
+        }
+        switch (node->txt.enc) {
+        default:
+        case 0:
+            n += node->txt.dim;
+            break;
+        case 1:
+            n += node->txt.dim; /* UCS2 -> Latin1 */
+            break;
+        }
+    }
+    return n;
+}
+
 static unsigned char *
 writeChars(unsigned char *frame, char const *str, size_t n)
 {
@@ -969,7 +1235,7 @@ writeChars(unsigned char *frame, char const *str, size_t n)
 }
 
 static unsigned char *
-writeUcs2s(unsigned char *frame, unsigned short *str, size_t n)
+writeUcs2s(unsigned char *frame, unsigned short const *str, size_t n)
 {
     while (n--) {
         *frame++ = 0xff & (*str >> 8);
@@ -978,12 +1244,26 @@ writeUcs2s(unsigned char *frame, unsigned short *str, size_t n)
     return frame;
 }
 
+static unsigned char *
+writeLoBytes(unsigned char *frame, unsigned short const *str, size_t n)
+{
+    str++; /* skip BOM */
+    while (n--) {
+        unsigned short c = *str++;
+        if (c < 0x20u || 0xff < c) {
+            c = 0x20; /* blank */
+        }
+        *frame++ = c;
+    }
+    return frame;
+}
+
 static unsigned char *
 set_frame_comment(unsigned char *frame, FrameDataNode const *node)
 {
     size_t const n = sizeOfCommentNode(node);
     if (n > 10) {
-        frame = set_4_byte_value(frame, ID_COMMENT);
+        frame = set_4_byte_value(frame, node->fid);
         frame = set_4_byte_value(frame, (uint32_t) (n - 10));
         /* clear 2-byte header flags */
         *frame++ = 0;
@@ -1027,6 +1307,17 @@ set_frame_custom2(unsigned char *frame, FrameDataNode const *node)
         *frame++ = 0;
         /* clear 1 encoding descriptor byte to indicate ISO-8859-1 format */
         *frame++ = node->txt.enc == 1 ? 1 : 0;
+        if (node->dsc.dim > 0) {
+            if (node->dsc.enc != 1) {
+                frame = writeChars(frame, node->dsc.ptr.l, node->dsc.dim);
+                *frame++ = 0;
+            }
+            else {
+                frame = writeUcs2s(frame, node->dsc.ptr.u, node->dsc.dim);
+                *frame++ = 0;
+                *frame++ = 0;
+            }
+        }
         if (node->txt.enc != 1) {
             frame = writeChars(frame, node->txt.ptr.l, node->txt.dim);
         }
@@ -1037,6 +1328,39 @@ set_frame_custom2(unsigned char *frame, FrameDataNode const *node)
     return frame;
 }
 
+static unsigned char *
+set_frame_wxxx(unsigned char *frame, FrameDataNode const *node)
+{
+    size_t const n = sizeOfWxxxNode(node);
+    if (n > 10) {
+        frame = set_4_byte_value(frame, node->fid);
+        frame = set_4_byte_value(frame, (unsigned long) (n - 10));
+        /* clear 2-byte header flags */
+        *frame++ = 0;
+        *frame++ = 0;
+        if (node->dsc.dim > 0) {
+            /* clear 1 encoding descriptor byte to indicate ISO-8859-1 format */
+            *frame++ = node->txt.enc == 1 ? 1 : 0;
+            if (node->dsc.enc != 1) {
+                frame = writeChars(frame, node->dsc.ptr.l, node->dsc.dim);
+                *frame++ = 0;
+            }
+            else {
+                frame = writeUcs2s(frame, node->dsc.ptr.u, node->dsc.dim);
+                *frame++ = 0;
+                *frame++ = 0;
+            }
+        }
+        if (node->txt.enc != 1) {
+            frame = writeChars(frame, node->txt.ptr.l, node->txt.dim);
+        }
+        else {
+            frame = writeLoBytes(frame, node->txt.ptr.u, node->txt.dim);
+        }
+    }
+    return frame;
+}
+
 static unsigned char *
 set_frame_apic(unsigned char *frame, const char *mimetype, const unsigned char *data, size_t size)
 {
@@ -1078,7 +1402,7 @@ id3tag_set_fieldvalue(lame_global_flags * gfp, const char *fieldvalue)
 {
     lame_internal_flags *gfc = gfp->internal_flags;
     if (fieldvalue && *fieldvalue) {
-        int const frame_id = toID3v2TagId(fieldvalue);
+        uint32_t const frame_id = toID3v2TagId(fieldvalue);
         char  **p = NULL;
         if (strlen(fieldvalue) < 5 || fieldvalue[4] != '=') {
             return -1;
@@ -1091,7 +1415,7 @@ id3tag_set_fieldvalue(lame_global_flags * gfp, const char *fieldvalue)
                     return -1;
                 }
                 gfc->tag_spec.values = (char **) p;
-                gfc->tag_spec.values[gfc->tag_spec.num_values++] = strdup(fieldvalue);
+                local_strdup(&gfc->tag_spec.values[gfc->tag_spec.num_values++], fieldvalue);
             }
         }
         gfc->tag_spec.flags |= CHANGED_FLAG;
@@ -1100,6 +1424,41 @@ id3tag_set_fieldvalue(lame_global_flags * gfp, const char *fieldvalue)
     return 0;
 }
 
+int
+id3tag_set_fieldvalue_utf16(lame_global_flags * gfp, const unsigned short *fieldvalue)
+{
+    if (fieldvalue && *fieldvalue) {
+        size_t dx = hasUcs2ByteOrderMarker(fieldvalue[0]);
+        char fid[5] = {0,0,0,0,0};
+        uint32_t const frame_id = toID3v2TagId_ucs2(fieldvalue+dx);
+        if (local_ucs2_strlen(fieldvalue) < (5+dx) || fieldvalue[dx+4] != '=') {
+            return -1;
+        }
+        fid[0] = (fieldvalue[dx+0] & 0x0ff);
+        fid[1] = (fieldvalue[dx+1] & 0x0ff);
+        fid[2] = (fieldvalue[dx+2] & 0x0ff);
+        fid[3] = (fieldvalue[dx+3] & 0x0ff);
+        if (frame_id != 0) {
+            unsigned short* txt = 0;
+            int     rc;
+            local_ucs2_substr(&txt, fieldvalue, dx+5, local_ucs2_strlen(fieldvalue));
+            rc = id3tag_set_textinfo_ucs2(gfp, fid, txt);
+            free(txt);
+            return rc;
+        }
+    }
+    return -1;
+}
+
+extern int
+id3tag_set_fieldvalue_ucs2(lame_global_flags * gfp, const unsigned short *fieldvalue);
+
+int
+id3tag_set_fieldvalue_ucs2(lame_global_flags * gfp, const unsigned short *fieldvalue)
+{
+    return id3tag_set_fieldvalue_utf16(gfp, fieldvalue);
+}
+
 size_t
 lame_get_id3v2_tag(lame_global_flags * gfp, unsigned char *buffer, size_t size)
 {
@@ -1111,18 +1470,28 @@ lame_get_id3v2_tag(lame_global_flags * gfp, unsigned char *buffer, size_t size)
     if (gfc == 0) {
         return 0;
     }
+    if (test_tag_spec_flags(gfc, V1_ONLY_FLAG)) {
+        return 0;
+    }
+#if 0
+    debug_tag_spec_flags(gfc, "lame_get_id3v2_tag");
+#endif
     {
+        int usev2 = test_tag_spec_flags(gfc, ADD_V2_FLAG | V2_ONLY_FLAG);
         /* calculate length of four fields which may not fit in verion 1 tag */
         size_t  title_length = gfc->tag_spec.title ? strlen(gfc->tag_spec.title) : 0;
         size_t  artist_length = gfc->tag_spec.artist ? strlen(gfc->tag_spec.artist) : 0;
         size_t  album_length = gfc->tag_spec.album ? strlen(gfc->tag_spec.album) : 0;
         size_t  comment_length = gfc->tag_spec.comment ? strlen(gfc->tag_spec.comment) : 0;
         /* write tag if explicitly requested or if fields overflow */
-        if ((gfc->tag_spec.flags & (ADD_V2_FLAG | V2_ONLY_FLAG))
-            || (title_length > 30)
-            || (artist_length > 30) || (album_length > 30)
+        if ((title_length > 30)
+            || (artist_length > 30)
+            || (album_length > 30)
             || (comment_length > 30)
             || (gfc->tag_spec.track_id3v1 && (comment_length > 28))) {
+            usev2 = 1;
+        }
+        if (usev2) {
             size_t  tag_size;
             unsigned char *p;
             size_t  adjusted_tag_size;
@@ -1132,7 +1501,9 @@ lame_get_id3v2_tag(lame_global_flags * gfp, unsigned char *buffer, size_t size)
             static const char *mime_png = "image/png";
             static const char *mime_gif = "image/gif";
 
-            id3v2AddAudioDuration(gfp);
+            if (gfp->num_samples != MAX_U_32_NUM) {
+                id3v2AddAudioDuration(gfc, gfp->num_samples);
+            }
 
             /* calulate size of tag starting with 10-byte tag header */
             tag_size = 10;
@@ -1151,28 +1522,30 @@ lame_get_id3v2_tag(lame_global_flags * gfp, unsigned char *buffer, size_t size)
                     albumart_mime = mime_gif;
                     break;
                 }
-                tag_size += 10 + 4 + strlen(albumart_mime) + gfc->tag_spec.albumart_size;
+                if (albumart_mime) {
+                    tag_size += 10 + 4 + strlen(albumart_mime) + gfc->tag_spec.albumart_size;
+                }
             }
             {
                 id3tag_spec *tag = &gfc->tag_spec;
                 if (tag->v2_head != 0) {
                     FrameDataNode *node;
                     for (node = tag->v2_head; node != 0; node = node->nxt) {
-                        switch (node->fid) {
-                        case ID_COMMENT:
+                        if (node->fid == ID_COMMENT) {
                             tag_size += sizeOfCommentNode(node);
-                            break;
-
-                        default:
+                        }
+                        else if (isFrameIdMatching(node->fid, FRAME_ID('W',0,0,0))) {
+                            tag_size += sizeOfWxxxNode(node);
+                        }
+                        else {
                             tag_size += sizeOfNode(node);
-                            break;
                         }
                     }
                 }
             }
-            if (gfc->tag_spec.flags & PAD_V2_FLAG) {
-                /* add 128 bytes of padding */
-                tag_size += 128;
+            if (test_tag_spec_flags(gfc, PAD_V2_FLAG)) {
+                /* add some bytes of padding */
+                tag_size += gfc->tag_spec.padding_size;
             }
             if (size < tag_size) {
                 return tag_size;
@@ -1215,14 +1588,14 @@ lame_get_id3v2_tag(lame_global_flags * gfp, unsigned char *buffer, size_t size)
                 if (tag->v2_head != 0) {
                     FrameDataNode *node;
                     for (node = tag->v2_head; node != 0; node = node->nxt) {
-                        switch (node->fid) {
-                        case ID_COMMENT:
+                        if (node->fid == ID_COMMENT) {
                             p = set_frame_comment(p, node);
-                            break;
-
-                        default:
+                        }
+                        else if (isFrameIdMatching(node->fid,FRAME_ID('W',0,0,0))) {
+                            p = set_frame_wxxx(p, node);
+                        }
+                        else {
                             p = set_frame_custom2(p, node);
-                            break;
                         }
                     }
                 }
@@ -1230,8 +1603,10 @@ lame_get_id3v2_tag(lame_global_flags * gfp, unsigned char *buffer, size_t size)
             for (i = 0; i < gfc->tag_spec.num_values; ++i) {
                 p = set_frame_custom(p, gfc->tag_spec.values[i]);
             }
-            p = set_frame_apic(p, albumart_mime, gfc->tag_spec.albumart,
-                               gfc->tag_spec.albumart_size);
+            if (albumart_mime) {
+                p = set_frame_apic(p, albumart_mime, gfc->tag_spec.albumart,
+                                   gfc->tag_spec.albumart_size);
+            }
             /* clear any padding bytes */
             memset(p, 0, tag_size - (p - buffer));
             return tag_size;
@@ -1244,8 +1619,13 @@ int
 id3tag_write_v2(lame_global_flags * gfp)
 {
     lame_internal_flags *gfc = gfp->internal_flags;
-    if ((gfc->tag_spec.flags & CHANGED_FLAG)
-        && !(gfc->tag_spec.flags & V1_ONLY_FLAG)) {
+#if 0
+    debug_tag_spec_flags(gfc, "write v2");
+#endif
+    if (test_tag_spec_flags(gfc, V1_ONLY_FLAG)) {
+        return 0;
+    }
+    if (test_tag_spec_flags(gfc, CHANGED_FLAG)) {
         unsigned char *tag = 0;
         size_t  tag_size, n;
 
@@ -1263,7 +1643,7 @@ id3tag_write_v2(lame_global_flags * gfp)
             size_t  i;
             /* write tag directly into bitstream at current position */
             for (i = 0; i < tag_size; ++i) {
-                add_dummy_byte(gfp, tag[i], 1);
+                add_dummy_byte(gfc, tag[i], 1);
             }
         }
         free(tag);
@@ -1305,10 +1685,12 @@ lame_get_id3v1_tag(lame_global_flags * gfp, unsigned char *buffer, size_t size)
     if (buffer == 0) {
         return 0;
     }
-    if ((gfc->tag_spec.flags & CHANGED_FLAG)
-        && !(gfc->tag_spec.flags & V2_ONLY_FLAG)) {
+    if (test_tag_spec_flags(gfc, V2_ONLY_FLAG)) {
+        return 0;
+    }
+    if (test_tag_spec_flags(gfc, CHANGED_FLAG)) {
         unsigned char *p = buffer;
-        int     pad = (gfc->tag_spec.flags & SPACE_V1_FLAG) ? ' ' : 0;
+        int     pad = test_tag_spec_flags(gfc, SPACE_V1_FLAG) ? ' ' : 0;
         char    year[5];
 
         /* set tag identifier */
@@ -1319,7 +1701,7 @@ lame_get_id3v1_tag(lame_global_flags * gfp, unsigned char *buffer, size_t size)
         p = set_text_field(p, gfc->tag_spec.title, 30, pad);
         p = set_text_field(p, gfc->tag_spec.artist, 30, pad);
         p = set_text_field(p, gfc->tag_spec.album, 30, pad);
-        snprintf(year, sizeof(year), "%d", gfc->tag_spec.year);
+        sprintf(year, "%d", gfc->tag_spec.year);
         p = set_text_field(p, gfc->tag_spec.year ? year : NULL, 4, pad);
         /* limit comment field to 28 bytes if a track is specified */
         p = set_text_field(p, gfc->tag_spec.comment, gfc->tag_spec.track_id3v1 ? 28 : 30, pad);
@@ -1337,6 +1719,7 @@ lame_get_id3v1_tag(lame_global_flags * gfp, unsigned char *buffer, size_t size)
 int
 id3tag_write_v1(lame_global_flags * gfp)
 {
+    lame_internal_flags *const gfc = gfp->internal_flags;
     size_t  i, n, m;
     unsigned char tag[128];
 
@@ -1347,7 +1730,7 @@ id3tag_write_v1(lame_global_flags * gfp)
     }
     /* write tag directly into bitstream at current position */
     for (i = 0; i < n; ++i) {
-        add_dummy_byte(gfp, tag[i], 1);
+        add_dummy_byte(gfc, tag[i], 1);
     }
     return (int) n;     /* ok, tag has fixed size of 128 bytes, well below 2GB */
 }
