@@ -1,7 +1,7 @@
 /*
- *	Machine dependent defines/includes for LAME.
+ *      Machine dependent defines/includes for LAME.
  *
- *	Copyright (c) 1999 A.L. Faber
+ *      Copyright (c) 1999 A.L. Faber
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -10,7 +10,7 @@
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Library General Public License for more details.
  *
  * You should have received a copy of the GNU Library General Public
@@ -22,7 +22,14 @@
 #ifndef LAME_MACHINE_H
 #define LAME_MACHINE_H
 
+#include "version.h"
+
+#if (LAME_RELEASE_VERSION == 0)
+#undef NDEBUG
+#endif
+
 #include <stdio.h>
+#include <assert.h>
 
 #ifdef STDC_HEADERS
 # include <stdlib.h>
@@ -32,7 +39,7 @@
 #  define strchr index
 #  define strrchr rindex
 # endif
-char *strchr (), *strrchr ();
+char   *strchr(), *strrchr();
 # ifndef HAVE_MEMCPY
 #  define memcpy(d, s, n) bcopy ((s), (d), (n))
 #  define memmove(d, s, n) bcopy ((s), (d), (n))
@@ -44,6 +51,7 @@ char *strchr (), *strrchr ();
 #else
 # include <math.h>
 #endif
+#include <limits.h>
 
 #include <ctype.h>
 
@@ -62,14 +70,26 @@ char *strchr (), *strrchr ();
 # include <sys/stat.h>
 #endif
 
-/* 
+#ifdef HAVE_INTTYPES_H
+# include <inttypes.h>
+#else
+# ifdef HAVE_STDINT_H
+#  include <stdint.h>
+# endif
+#endif
+
+#ifdef WITH_DMALLOC
+#include <dmalloc.h>
+#endif
+
+/*
  * 3 different types of pow() functions:
  *   - table lookup
  *   - pow()
  *   - exp()   on some machines this is claimed to be faster than pow()
  */
 
-#define POW20(x) (assert(-Q_MAX2 <= x && x < Q_MAX), pow20[x+Q_MAX2])
+#define POW20(x) (assert(0 <= (x+Q_MAX2) && x < Q_MAX), pow20[x+Q_MAX2])
 /*#define POW20(x)  pow(2.0,((double)(x)-210)*.25) */
 /*#define POW20(x)  exp( ((double)(x)-210)*(.25*LOG2) ) */
 
@@ -112,7 +132,7 @@ char *strchr (), *strrchr ();
 # define FLOAT_MAX FLT_MAX
 #else
 # ifndef FLOAT
-typedef float   FLOAT;
+typedef float FLOAT;
 #  ifdef FLT_MAX
 #   define FLOAT_MAX FLT_MAX
 #  else
@@ -121,8 +141,8 @@ typedef float   FLOAT;
 # endif
 #endif
 
-#ifndef FLOAT8  /* NOTE: RH: 7/00:  if FLOAT8=float, it breaks resampling and VBR code */
-typedef double  FLOAT8;
+#ifndef FLOAT8
+typedef double FLOAT8;
 # ifdef DBL_MAX
 #  define FLOAT8_MAX DBL_MAX
 # else
@@ -137,10 +157,22 @@ typedef double  FLOAT8;
 #endif
 
 /* sample_t must be floating point, at least 32 bits */
-typedef FLOAT     sample_t;
-typedef sample_t  stereo_t [2];
+typedef FLOAT sample_t;
+
+#define dimension_of(array) (sizeof(array)/sizeof(array[0]))
+#define beyond(array) (array+dimension_of(array))
+
+#if 1
+#define EQ(a,b) (\
+(fabs(a) > fabs(b)) \
+ ? (fabs((a)-(b)) <= (fabs(a) * 1e-6f)) \
+ : (fabs((a)-(b)) <= (fabs(b) * 1e-6f)))
+#else
+#define EQ(a,b) (fabs((a)-(b))<1E-37)
+#endif
+
+#define NEQ(a,b) (!EQ(a,b))
 
 #endif
 
 /* end of machine.h */
-
